// WmaAlbumArtist.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "LeoHelpers.h"

class BinaryData
{
protected:
	BYTE *m_pData;
	DWORD m_dwSize;

public:
	BinaryData()
		: m_pData(NULL)
		, m_dwSize(0)
	{
	}

	~BinaryData()
	{
		Clear();
	}

	void Clear()
	{
		delete [] m_pData;
		m_pData = NULL;
		m_dwSize = 0;
	}

	void SetAndOwn(BYTE *pData, DWORD dwSize)
	{
		Clear();

		m_pData = pData;
		m_dwSize = dwSize;
	}

	bool Load(const TCHAR *filename)
	{
		bool bResult = false;

		delete [] m_pData;
		m_pData = NULL;
		m_dwSize = 0;

		HANDLE hFile = ::CreateFile(filename, FILE_READ_ATTRIBUTES|FILE_READ_DATA, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);

		if (INVALID_HANDLE_VALUE == hFile)
		{
			_tcprintf(_T("Failed to open %s\n"), filename);
		}
		else
		{
			DWORD dwFileSize = ::GetFileSize(hFile, NULL);

			if (INVALID_FILE_SIZE == dwFileSize)
			{
				_tcprintf(_T("Failed to get size of %s\n"), filename);
			}
			else if (0 == dwFileSize)
			{
				_tcprintf(_T("%s is empty\n"), filename);
			}
			else
			{
				m_pData = new(std::nothrow) BYTE[dwFileSize];

				if (m_pData == NULL)
				{
					_tcprintf(_T("Failed to allocate memory for %s\n"), filename);
				}
				else
				{
					if (0 == ::ReadFile(hFile, m_pData, dwFileSize, &m_dwSize, NULL) || m_dwSize != dwFileSize)
					{
						_tcprintf(_T("Failed to read %s\n"), filename);
					}
					else
					{
						bResult = true;
					}
				}
			}

			::CloseHandle(hFile);
		}

		if (!bResult)
		{
			delete [] m_pData;
			m_pData = NULL;
			m_dwSize = 0;
		}

		return bResult;
	}

	bool Save(const TCHAR *filename)
	{
		return Save(filename, m_pData, m_dwSize);
	}

	static bool Save(const TCHAR *filename, const BYTE *pData, DWORD dwSize)
	{
		bool bResult = false;

		HANDLE hFile = ::CreateFile(filename, FILE_WRITE_DATA, FILE_SHARE_READ, NULL, CREATE_NEW, FILE_FLAG_SEQUENTIAL_SCAN, NULL);

		if (hFile == INVALID_HANDLE_VALUE)
		{
			DWORD dwErr = GetLastError();

			if (dwErr == ERROR_ALREADY_EXISTS
			||	dwErr == ERROR_FILE_EXISTS)
			{
				_tcprintf(_T("Failed to save (already exists): %s\n"), filename);
			}
			else
			{
				_tcprintf(_T("Failed to save (error %ld): %s\n"), dwErr, filename);
			}
		}
		else
		{
			DWORD dwWritten = 0;

			if (!WriteFile(hFile, pData, dwSize, &dwWritten, NULL))
			{
				DWORD dwErr = GetLastError();
				_tcprintf(_T("Failed to write (error %ld): %s\n"), dwErr, filename);
			}
			else if (dwWritten != dwSize)
			{
				_tcprintf(_T("Failed to write (only partially written): %s\n"), filename);
			}
			else
			{
				bResult = true;
			}

			CloseHandle(hFile);

			if (!bResult)
			{
				DeleteFile(filename);
			}
		}

		return bResult;
	}

	BYTE *GetBuffer()
	{
		return m_pData;
	}

	DWORD GetSize()
	{
		return m_dwSize;
	}
};

// Well, it sucks more in some ways but it sucks less for our purposes.
// Wraps a small part of the Windows Media API so that we don't have to worry about memory allocation
// (calling functions twice to find out how much memory to allocate is *so* 1980s and it was
// stupid even back then!) and COM rubbish.
class SuckLessWmMetadataEditor
{
protected:
	IWMMetadataEditor *m_pEditor;
	IWMHeaderInfo3 *m_pHeader;
	bool m_bClose;
	std::basic_string<TCHAR> m_currentFilename;

public:
	SuckLessWmMetadataEditor()
		: m_pEditor(NULL)
		, m_pHeader(NULL)
		, m_bClose(false)
	{
	}

	~SuckLessWmMetadataEditor()
	{
		if (m_bClose && m_pEditor != NULL)
		{
			m_pEditor->Close();
			m_bClose = false;
		}

		if (m_pHeader != NULL)
		{
			m_pHeader->Release();
			m_pHeader = NULL;
		}

		if (m_pEditor != NULL)
		{
			m_pEditor->Release();
			m_pEditor = NULL;
		}
	}

	bool Init()
	{
		bool bResult = false;

		if (m_pEditor != NULL || m_pHeader != NULL)
		{
			_tcprintf(_T("SuckLessWmMetadataEditor::Init called twice\n"));
		}
		else
		{
			HRESULT hr = WMCreateEditor(&m_pEditor);

			if (hr != S_OK || m_pEditor == NULL)
			{
				m_pEditor = NULL;
				_tcprintf(_T("Failed to create editor\n"));
			}
			else
			{
				hr = m_pEditor->QueryInterface(__uuidof(IWMHeaderInfo3), reinterpret_cast<void**>(&m_pHeader));
				if (hr != S_OK || m_pHeader == NULL)
				{
					m_pHeader = NULL;
					_tcprintf(_T("Failed to get WMA Header object\n"));
				}
				else
				{
					bResult = true;
				}
			}
		}

		return bResult;
	}

	bool Open(LPCTSTR filename)
	{
		bool bResult = false;

		if (m_pEditor == NULL || m_pHeader == NULL)
		{
			_tcprintf(_T("SuckLessWmMetadataEditor::Init was not done\n"));
		}
		else
		{
			if (m_bClose)
			{
				m_pEditor->Close();
				m_bClose = false;
				m_currentFilename.clear();
			}

			HRESULT hr = m_pEditor->Open(filename);

			if (hr != S_OK)
			{
				_tcprintf(_T("Failed to open %s\n"), filename);
			}
			else
			{
				m_currentFilename = filename;
				m_bClose = true;
				bResult = true;
			}
		}

		return bResult;
	}

	bool Close(bool saveChanges)
	{
		bool bResult = false;

		if (m_pEditor == NULL || m_pHeader == NULL)
		{
			_tcprintf(_T("SuckLessWmMetadataEditor::Init was not done\n"));
		}
		else if (m_bClose)
		{
			HRESULT hr;

			if (saveChanges)
			{
				hr = m_pEditor->Flush();
			}
			else
			{
				hr = m_pEditor->Close();
			}

			if (hr != S_OK)
			{
				_tcprintf(_T("Failed to %s %s\n"), saveChanges ? _T("flush") : _T("close"), m_currentFilename.c_str());
			}
			else
			{
				bResult = true;
			}

			m_bClose = false;
		}

		return bResult;
	}

	bool GetBinaryAttribute(LPCTSTR attributeName, bool *pbExists, BinaryData *pBinData)
	{
		bool bResult = false;

		if (pbExists != NULL)
		{
			*pbExists = false;
		}

		if (pBinData != NULL)
		{
			pBinData->Clear();
		}

		WMT_ATTR_DATATYPE type = WMT_TYPE_BINARY;
		BYTE *pValue = NULL;
		WORD length = 0;

		if (getAttributeByName(attributeName, &type, &pValue, &length))
		{
			if (pValue == NULL)
			{
				bResult = true;
			}
			else
			{
				if (type != WMT_TYPE_BINARY)
				{
					_tcprintf(_T("Attribute %s is not binary in %s\n"), attributeName, m_currentFilename.c_str());
				}
				else
				{
					if (pBinData != NULL)
					{
						pBinData->SetAndOwn(pValue, length);
						pValue = NULL; // Prevent deletion below.
					}

					if (pbExists != NULL)
					{
						*pbExists = true;
					}

					bResult = true;
				}

				delete [] pValue;
			}
		}

		return bResult;
	}

	bool GetBinaryAttributeByIndex(WORD wIndex, bool *pbExists, BinaryData *pBinData, std::basic_string<TCHAR> *pstrAttributeName)
	{
		bool bResult = false;

		if (pbExists != NULL)
		{
			*pbExists = false;
		}

		if (pBinData != NULL)
		{
			pBinData->Clear();
		}

		if (pstrAttributeName != NULL)
		{
			pstrAttributeName->clear();
		}

		WMT_ATTR_DATATYPE type = WMT_TYPE_BINARY;
		BYTE *pValue = NULL;
		DWORD length = 0;

		WCHAR *foundAttributeName = NULL;

		if (getAttributeByIndex(wIndex, &foundAttributeName, &type, &pValue, &length))
		{
			if (pValue == NULL)
			{
				bResult = true;
			}
			else
			{
				if (type != WMT_TYPE_BINARY)
				{
					_tcprintf(_T("Attribute #%hd is not binary in %s\n"), wIndex, m_currentFilename.c_str());
				}
				else
				{
					if (pBinData != NULL)
					{
						pBinData->SetAndOwn(pValue, length);
						pValue = NULL; // Prevent deletion below.
					}

					if (pbExists != NULL)
					{
						*pbExists = true;
					}

					if (foundAttributeName != NULL && pstrAttributeName != NULL)
					{
						*pstrAttributeName = foundAttributeName;
					}

					bResult = true;
				}
			}

			delete [] pValue;
			delete [] foundAttributeName;
		}

		return bResult;
	}

	bool GetStringAttribute(LPCTSTR attributeName, bool *pbExists, std::basic_string<TCHAR> *pString)
	{
		bool bResult = false;

		if (pbExists != NULL)
		{
			*pbExists = false;
		}

		if (pString != NULL)
		{
			pString->clear();
		}

		WMT_ATTR_DATATYPE type = WMT_TYPE_DWORD;
		BYTE *pValue = NULL;
		WORD length = 0;

		if (getAttributeByName(attributeName, &type, &pValue, &length))
		{
			if (pValue == NULL)
			{
				bResult = true;
			}
			else
			{
				if (type != WMT_TYPE_STRING)
				{
					_tcprintf(_T("Attribute %s is not a string in %s\n"), attributeName, m_currentFilename.c_str());
				}
				else
				{
					if (pString != NULL)
					{
						*pString = reinterpret_cast<LPCWSTR>(pValue);
					}

					if (pbExists != NULL)
					{
						*pbExists = true;
					}

					bResult = true;
				}

				delete[] pValue;
			}
		}

		return bResult;
	}

	bool AddBinaryAttribute(LPCTSTR attributeName, const BYTE *pData, DWORD dwLength)
	{
		bool bResult = false;

		if (addAttribute(attributeName, WMT_TYPE_BINARY, pData, dwLength))
		{
			bResult = true;
		}

		return bResult;
	}

	bool AddStringAttribute(LPCTSTR attributeName, const TCHAR *szAttributeValue)
	{
		bool bResult = false;

		const WCHAR *wszValue = szAttributeValue;
		DWORD dwLength = (wcslen(wszValue) + 1) * sizeof(WCHAR);

		if (addAttribute(attributeName, WMT_TYPE_STRING, reinterpret_cast<const BYTE *>(wszValue), dwLength))
		{
			bResult = true;
		}

		return bResult;
	}

	bool DeleteAttribute(LPCTSTR attributeName)
	{
		bool bResult = false;

		WORD *pwIndices;
		WORD wIndexCount;

		if (getAttributeIndices(attributeName, &pwIndices, &wIndexCount))
		{
			bResult = true;

			for(WORD wi = 0; wi < wIndexCount; ++wi)
			{
				if (!deleteAttribute(attributeName, pwIndices[wi]))
				{
					bResult = false;
					break;
				}
			}

			delete [] pwIndices;
		}

		return bResult;
	}

	bool GetAttributeIndices(LPCTSTR attributeName, std::vector<WORD> *pvecIndices)
	{
		bool bResult = false;

		if (pvecIndices != NULL)
		{
			pvecIndices->clear();
		}

		WORD *pwIndices;
		WORD wIndexCount;

		if (getAttributeIndices(attributeName, &pwIndices, &wIndexCount))
		{
			if (wIndexCount > 0)
			{
				bResult = true;

				if (pvecIndices != NULL)
				{
					pvecIndices->reserve(wIndexCount);

					for (WORD wi = 0; wi < wIndexCount; ++wi)
					{
						pvecIndices->push_back(pwIndices[wi]);
					}
				}
			}

			delete [] pwIndices;
		}

		return bResult;
	}

protected:

	bool getAttributeIndices(LPCTSTR attributeName, WORD **ppwOutIndices, WORD *pwOutIndexCount)
	{
		bool bResult = false;
		*pwOutIndexCount = 0;
		*ppwOutIndices = NULL;

		if (m_pEditor == NULL || m_pHeader == NULL)
		{
			_tcprintf(_T("SuckLessWmMetadataEditor::Init was not done\n"));
		}
		else if (!m_bClose)
		{
			_tcprintf(_T("Cannot get attribute when no file is open\n"));
		}
		else
		{
			HRESULT hr;

			hr = m_pHeader->GetAttributeIndices(0, attributeName, NULL, NULL, pwOutIndexCount);

			if (hr != S_OK)
			{
				_tcprintf(_T("Failed to get %s attribute from %s\n"), attributeName, m_currentFilename.c_str());
			}
			else if (*pwOutIndexCount == 0)
			{
				_tcprintf(_T("No %s attribute in %s\n"), attributeName, m_currentFilename.c_str());
			}
			else
			{
				*ppwOutIndices = new WORD[*pwOutIndexCount];

				hr = m_pHeader->GetAttributeIndices(0, attributeName, NULL, *ppwOutIndices, pwOutIndexCount);

				if (hr != S_OK)
				{
					delete [] *ppwOutIndices;
					*ppwOutIndices = NULL;

					_tcprintf(_T("Failed to get %s attribute from %s\n"), attributeName, m_currentFilename.c_str());
				}
				else
				{
					bResult = true;
				}
			}
		}

		return bResult;
	}

	bool deleteAttribute(LPCTSTR attributeName, WORD wIndex)
	{
		bool bResult = false;

		if (m_pEditor == NULL || m_pHeader == NULL)
		{
			_tcprintf(_T("SuckLessWmMetadataEditor::Init was not done\n"));
		}
		else if (!m_bClose)
		{
			_tcprintf(_T("Cannot delete attribute when no file is open\n"));
		}
		else
		{
			HRESULT hr;
			
			hr = m_pHeader->DeleteAttribute(0, wIndex);

			if (hr != S_OK)
			{
				_tcprintf(_T("Failed to delete %s attribute from %s\n"), attributeName, m_currentFilename.c_str());
			}
			else
			{
				bResult = true;
			}
		}

		return bResult;
	}

	// Succeeds and returns NULL if the attribute doesn't exist on the file.
	bool getAttributeByIndex(WORD wIndex, WCHAR **ppStrOutName, WMT_ATTR_DATATYPE *pOutType, BYTE **ppOutValue, DWORD *pcbOutLength)
	{
		bool bResult = false;
		*ppStrOutName = NULL;
		*pcbOutLength = 0;
		*ppOutValue = NULL;

		if (m_pEditor == NULL || m_pHeader == NULL)
		{
			_tcprintf(_T("SuckLessWmMetadataEditor::Init was not done\n"));
		}
		else if (!m_bClose)
		{
			_tcprintf(_T("Cannot get attribute when no file is open\n"));
		}
		else
		{
			WORD streamNum = 0;

			HRESULT hr;

			WORD wNameLen = 0;
			WORD wLangIdx = 0;

			hr = m_pHeader->GetAttributeByIndexEx(streamNum, wIndex, NULL, &wNameLen, pOutType, &wLangIdx, NULL, pcbOutLength);
			if (hr == ASF_E_NOTFOUND)
			{
				bResult = true;
			}
			else if (hr != S_OK)
			{
				_tcprintf(_T("Failed to get size of attribute #%hd from %s\n"), wIndex, m_currentFilename.c_str());
			}
			else
			{
				*ppStrOutName = new(std::nothrow) WCHAR[wNameLen];
				*ppOutValue = new(std::nothrow) BYTE[*pcbOutLength];

				if (*ppStrOutName == NULL
				||	*ppOutValue == NULL)
				{
					delete [] *ppStrOutName;
					delete [] *ppOutValue;
					*ppStrOutName = NULL;
					*ppOutValue = NULL;

					_tcprintf(_T("Failed to allocate memory for attribute #%hd from %s\n"), wIndex, m_currentFilename.c_str());
				}
				else
				{
					hr = m_pHeader->GetAttributeByIndexEx(streamNum, wIndex, *ppStrOutName, &wNameLen, pOutType, &wLangIdx, *ppOutValue, pcbOutLength);
					if (hr != S_OK)
					{
						delete [] *ppStrOutName;
						delete [] *ppOutValue;
						*ppStrOutName = NULL;
						*ppOutValue = NULL;

						_tcprintf(_T("Failed to get attribute #%hd from %s\n"), wIndex, m_currentFilename.c_str());
					}
					else
					{
						bResult = true;
					}
				}
			}
		}

		return bResult;
	}

	// Succeeds and returns NULL if the attribute doesn't exist on the file.
	bool getAttributeByName(LPCTSTR attributeName, WMT_ATTR_DATATYPE *pOutType, BYTE **ppOutValue, WORD *pcbOutLength)
	{
		bool bResult = false;
		*pcbOutLength = 0;
		*ppOutValue = NULL;

		if (m_pEditor == NULL || m_pHeader == NULL)
		{
			_tcprintf(_T("SuckLessWmMetadataEditor::Init was not done\n"));
		}
		else if (!m_bClose)
		{
			_tcprintf(_T("Cannot get attribute when no file is open\n"));
		}
		else
		{
			WORD streamNum = 0;

			HRESULT hr;

			hr = m_pHeader->GetAttributeByName(&streamNum, attributeName, pOutType, NULL, pcbOutLength);
			if (hr == ASF_E_NOTFOUND)
			{
				bResult = true;
			}
			else if (hr != S_OK)
			{
				_tcprintf(_T("Failed to get size of %s attribute from %s\n"), attributeName, m_currentFilename.c_str());
			}
			else
			{
				*ppOutValue = new(std::nothrow) BYTE[*pcbOutLength];

				if (*ppOutValue == NULL)
				{
					_tcprintf(_T("Failed to allocate memory for %s attribute from %s\n"), attributeName, m_currentFilename.c_str());
				}
				else
				{
					hr = m_pHeader->GetAttributeByName(&streamNum, attributeName, pOutType, *ppOutValue, pcbOutLength);
					if (hr != S_OK)
					{
						delete [] *ppOutValue;
						*ppOutValue = NULL;

						_tcprintf(_T("Failed to get %s attribute from %s\n"), attributeName, m_currentFilename.c_str());
					}
					else
					{
						bResult = true;
					}
				}
			}
		}

		return bResult;
	}

	bool addAttribute(LPCTSTR attributeName, WMT_ATTR_DATATYPE type, const BYTE *pValue, DWORD dwLength)
	{
		bool bResult = false;

		if (m_pEditor == NULL || m_pHeader == NULL)
		{
			_tcprintf(_T("SuckLessWmMetadataEditor::Init was not done\n"));
		}
		else if (!m_bClose)
		{
			_tcprintf(_T("Cannot add attribute when no file is open\n"));
		}
		else
		{
			WORD wIdx = 0;

			HRESULT hr = m_pHeader->AddAttribute(0, attributeName, &wIdx, type, 0, pValue, dwLength);
			if (hr != S_OK)
			{
				_tcprintf(_T("Failed to add %s attribute to %s\n"), attributeName, m_currentFilename.c_str());
			}
			else
			{
				bResult = true;
			}
		}

		return bResult;
	}

};

bool parentStartsWithPrefix(const TCHAR *filename, const TCHAR *prefix, const TCHAR *exact, bool bAnyParent)
{
	TCHAR *path = NLeoHelpers::GetParentPath(filename);

	const size_t lenPrefix = _tcslen(prefix);

	while (path != NULL)
	{
		if (path[0] == L'\0')
		{
			delete[] path;
			return false;
		}

		const TCHAR *lastPart = NLeoHelpers::GetLastPathPart(path);

		if (0 == _tcsnicmp(prefix, lastPart, lenPrefix)
		||	0 == _tcsicmp(exact,   lastPart))
		{
			delete[] path;
			return true;
		}

		if (!bAnyParent)
		{
			delete[] path;
			return false;
		}

		TCHAR *path2 = NLeoHelpers::GetParentPath(path);
		delete[] path;
		path = path2;
	}

	return false;
}

int _tmain(int argc, _TCHAR* argv[])
{
	// World's worst command-line parsing comin' right up! :-)

	bool bSetAA = false;
	bool bSetAAVA = false;
	bool bSetAAVV = false;
	bool bClearAA = false;
	bool bSetCover = false;
	bool bClearCover = false;
	bool bSaveCover = false;
	bool bHelp = false;

	TCHAR fullPath1[MAX_PATH] = {0};

	const TCHAR *coverFileName = NULL;
	const TCHAR *wmaFileName1 = NULL;

	if (argc >= 3 && 0 == _tcsicmp(_T("/SetAA"), argv[1]))
	{
		bSetAA = true;

		int argi = 2;

		while(argi < (argc-1))
		{
			if (0 == _tcsicmp(_T("/VA"), argv[argi]))
			{
				bSetAAVA = true;
			}
			else if (0 == _tcsicmp(_T("/VV"), argv[argi]))
			{
				bSetAAVV = true;
			}
			else
			{
				bHelp = true;
				break;
			}

			++argi;
		}

		if (!bHelp)
		{
			wmaFileName1 = argv[argi];
		}
	}
	else if (argc == 3 && 0 == _tcsicmp(_T("/ClearAA"), argv[1]))
	{
		bClearAA = true;
		wmaFileName1 = argv[2];
	}
	else if (argc == 4 && 0 == _tcsicmp(_T("/SetCover"), argv[1]))
	{
		bSetCover = true;
		coverFileName = argv[2];
		wmaFileName1 = argv[3];
	}
	else if (argc == 3 && 0 == _tcsicmp(_T("/ClearCover"), argv[1]))
	{
		bClearCover = true;
		wmaFileName1 = argv[2];
	}
	else if (argc == 4 && 0 == _tcsicmp(_T("/SaveCover"), argv[1]))
	{
		bSaveCover = true;
		wmaFileName1 = argv[2];
		coverFileName = argv[3];
	}
	else
	{
		bHelp = true;
	}

	if (wmaFileName1 != NULL)
	{
		DWORD dwRes = GetFullPathName(wmaFileName1, _countof(fullPath1), fullPath1, NULL);

		if (dwRes != 0 && dwRes < _countof(fullPath1))
		{
			wmaFileName1 = fullPath1;
		}
	}

	if (bHelp)
	{
		_tcprintf(_T("\nTo set the AlbumArtist field:\n"));
		_tcprintf(_T("  WmaAlbumArtist /SetAA [/VA] [/VV] <musicFilePath>\n"));
		_tcprintf(_T("\nTo clear the AlbumArtist field:\n"));
		_tcprintf(_T("  WmaAlbumArtist /ClearAA <musicFilePath>\n"));
		_tcprintf(_T("\nTo set the Cover Image:\n"));
		_tcprintf(_T("  WmaAlbumArtist /SetCover <jpegFileName> <musicFilePath>\n"));
		_tcprintf(_T("\nTo clear the Cover Image:\n"));
		_tcprintf(_T("  WmaAlbumArtist /ClearCover <musicFilePath>\n"));
		_tcprintf(_T("\nTo save the Cover Image to a file:\n"));
		_tcprintf(_T("  WmaAlbumArtist /SaveCover <musicFilePath> <imageFileNameNoExtension>\n"));
		_tcprintf(_T("\nGive only one WMA/MP3 filename.\n\nPassing directories or multiple files isn't supported, for now at least.\nUse a batch file or a configurable file manager like Directory Opus (www.dopus.com) to automate the tool.\n"));
		_tcprintf(_T("\nThe optional /VA arg makes the /SetAA command use \"Various Artists\" instead of \"Various\".\n"));
		_tcprintf(_T("\nThe optional /VV arg makes the /SetAA command look for any parent/ancestor folder starting with \"Various\", not just the direct parent.\n"));
		_tcprintf(_T("\nThe cover filenames should be names only, without paths. They will be loaded/saved in the same folder as the WMA/MP3 file.\n"));
		_tcprintf(_T("\nC++ Source Code available from http://www.PretentiousName.com\n"));
	}
	else
	{
		HRESULT hr = CoInitializeEx(NULL, COINIT_MULTITHREADED);

		if (hr != S_OK)
		{
			_tcprintf(_T("Failed to initialize COM\n"));
		}
		else
		{
			SuckLessWmMetadataEditor ed;

			if (ed.Init())
			{
				if (ed.Open(wmaFileName1))
				{
					bool saveChanges = false;

					if (bSetAA)
					{
						bool bVarious = parentStartsWithPrefix(wmaFileName1, _T("Various "), _T("Various"), bSetAAVV);

						bool bAlbumArtistExists = false;
						std::basic_string<TCHAR> strAlbumArtist;

						if (ed.GetStringAttribute(g_wszWMAlbumArtist, &bAlbumArtistExists, &strAlbumArtist))
						{
							bool bArtistExists = false;
							std::basic_string<TCHAR> strArtist;

							if (bAlbumArtistExists)
							{
								_tcprintf(_T("Skipping (already has AlbumArtist=\"%s\"): %s\n"), strAlbumArtist.c_str(), wmaFileName1);
							}
							else if (bVarious)
							{
								const TCHAR *variousInsert = bSetAAVA ? _T("Various Artists") : _T("Various");

								if (ed.AddStringAttribute(g_wszWMAlbumArtist, variousInsert))
								{
									saveChanges = true;
									_tcprintf(_T("Set AlbumArtist=\"%s\": %s\n"), variousInsert, wmaFileName1);
								}
							}
							else if (ed.GetStringAttribute(g_wszWMAuthor, &bArtistExists, &strArtist))
							{
								if (!bArtistExists)
								{
									_tcprintf(_T("Skipping (has no artist!): %s\n"), wmaFileName1);
								}
								else
								{
									if (ed.AddStringAttribute(g_wszWMAlbumArtist, strArtist.c_str()))
									{
										saveChanges = true;
										_tcprintf(_T("Set AlbumArtist=\"%s\": %s\n"), strArtist.c_str(), wmaFileName1);
									}
								}
							}
						}
					}
					else if (bClearAA)
					{
						bool bAlbumArtistExists = false;
						std::basic_string<TCHAR> strAlbumArtist;

						if (ed.GetStringAttribute(g_wszWMAlbumArtist, &bAlbumArtistExists, &strAlbumArtist))
						{
							if (!bAlbumArtistExists)
							{
								_tcprintf(_T("Skipping (has no AlbumArtist): %s\n"), wmaFileName1);
							}
							else
							{
								if (ed.DeleteAttribute(g_wszWMAlbumArtist))
								{
									saveChanges = true;
									_tcprintf(_T("Cleared AlbumArtist: %s\n"), wmaFileName1);
								}
							}
						}
					}
					else if (bClearCover)
					{
						if (ed.DeleteAttribute(g_wszWMPicture))
						{
							saveChanges = true;
							_tcprintf(_T("Removed picture: %s\n"), wmaFileName1);
						}
					}
					else if (bSetCover)
					{
						TCHAR *parentPath = NLeoHelpers::GetParentPath(wmaFileName1);

						if (parentPath == NULL)
						{
							_tcprintf(_T("Couldn't get parent of %s\n"), wmaFileName1);
						}
						else
						{
							TCHAR *coverPath = NLeoHelpers::JoinPaths(parentPath, coverFileName);

							if (coverPath == NULL)
							{
								_tcprintf(_T("Couldn't add %s to %s\n"), coverFileName, parentPath);
							}
							else
							{
								bool bPictureExists = false;

								if (ed.GetBinaryAttribute(g_wszWMPicture, &bPictureExists, NULL))
								{
									if (bPictureExists)
									{
										_tcprintf(_T("Skipping (already has picture): %s\n"), wmaFileName1);
									}
									else
									{
										BinaryData bd;
										
										if (bd.Load(coverPath))
										{
											WM_PICTURE wmPicture;
											ZeroMemory(&wmPicture, sizeof(WM_PICTURE));
											wmPicture.pwszMIMEType = L"image/jpeg";
											wmPicture.bPictureType = 3; // Front album cover.
											wmPicture.pwszDescription = L"";
											wmPicture.dwDataLen = bd.GetSize();
											wmPicture.pbData = bd.GetBuffer();

											if (ed.AddBinaryAttribute(g_wszWMPicture, reinterpret_cast<BYTE *>(&wmPicture), sizeof(WM_PICTURE)))
											{
												_tcprintf(_T("Added picture: %s\n"), wmaFileName1);

												saveChanges = true;
											}
										}
									}
								}

								delete [] coverPath;
							}

							delete [] parentPath;
						}
					}
					else if (bSaveCover)
					{
						TCHAR *parentPath = NLeoHelpers::GetParentPath(wmaFileName1);

						if (parentPath == NULL)
						{
							_tcprintf(_T("Couldn't get parent of %s\n"), wmaFileName1);
						}
						else
						{
							TCHAR *coverPath = NLeoHelpers::JoinPaths(parentPath, coverFileName);

							if (coverPath == NULL)
							{
								_tcprintf(_T("Couldn't add %s to %s\n"), coverFileName, parentPath);
							}
							else
							{
								bool bPictureExists = false;
								BinaryData bd;

								std::vector<WORD> vecIndices;

								if (!ed.GetAttributeIndices(g_wszWMPicture, &vecIndices))
								{
									_tcprintf(_T("Skipping (has no picture): %s\n"), wmaFileName1);
								}
								else
								{
									for(unsigned int i = 0; i < vecIndices.size(); ++i)
									{
										if (ed.GetBinaryAttributeByIndex(vecIndices[i], &bPictureExists, &bd, NULL))
										{
											if (!bPictureExists)
											{
												_tcprintf(_T("Skipping (has no picture): %s\n"), wmaFileName1);
											}
											else
											{
												WM_PICTURE *pwmPicture = reinterpret_cast<WM_PICTURE *>(bd.GetBuffer());

												const wchar_t *szMimeType = pwmPicture->pwszMIMEType;
												int iPicType = pwmPicture->bPictureType;

												std::basic_string<TCHAR> strRegPath = _T("MIME\\Database\\Content Type\\");
												strRegPath += szMimeType;

												std::basic_string<TCHAR> strExtension;

												// Find the extension from its MIME type registry settings.
												// If the MIME type isn't in the registry and it's image/jpeg, also try image/jpg in case that's registered instead.
												// If the MIME type isn't in the registry and it's image/jpg, also try image/jpeg in case that's registered instead.

												if (!NLeoHelpers::LeetRegQueryStringValue(HKEY_CLASSES_ROOT, strRegPath.c_str(), _T("Extension"), 0, &strExtension)
												&&	(0 != _wcsicmp(szMimeType, L"image/jpg")
												||	 !NLeoHelpers::LeetRegQueryStringValue(HKEY_CLASSES_ROOT, L"MIME\\Database\\Content Type\\image/jpeg", _T("Extension"), 0, &strExtension))
												&&	(0 != _wcsicmp(szMimeType, L"image/jpeg")
												||	 !NLeoHelpers::LeetRegQueryStringValue(HKEY_CLASSES_ROOT, L"MIME\\Database\\Content Type\\image/jpg", _T("Extension"), 0, &strExtension)))
												{
													_tcprintf(_T("Skipping (picture has unknown MIME type \"%s\"): %s\n"), szMimeType, wmaFileName1);
												}
												else
												{
													TCHAR *szFullCoverPath;
													
													if (vecIndices.size() == 1)
													{
														szFullCoverPath = NLeoHelpers::StringAllocAndFormat(_T("%s%s"), coverPath, strExtension.c_str());
													}
													else
													{
														szFullCoverPath = NLeoHelpers::StringAllocAndFormat(_T("%s_%u%s"), coverPath, i, strExtension.c_str());
													}

													if (szFullCoverPath == NULL)
													{
														_tcprintf(_T("Skipping (failed to build path): %s\n"), wmaFileName1);
													}
													else
													{
														if (bd.Save(szFullCoverPath, pwmPicture->pbData, pwmPicture->dwDataLen))
														{
															if (pwmPicture->pwszDescription != NULL
															&&	pwmPicture->pwszDescription[0] != L'\0')
															{
																_tcprintf(_T("Saved picture pictype %d (%s): %s\n"), iPicType, pwmPicture->pwszDescription, szFullCoverPath);
															}
															else
															{
																_tcprintf(_T("Saved picture pictype %d: %s\n"), iPicType, szFullCoverPath);
															}
														}

														delete [] szFullCoverPath;
													}
												}
											}
										}
									}
								}

								delete [] coverPath;
							}

							delete [] parentPath;
						}
					}

					ed.Close(saveChanges);
				}
			}

			CoUninitialize();
		}
	}

	return 0;
}
